@@ -42,11 +42,11 @@ MediaWiki::API - Provides a Perl interface to the MediaWiki API (http://www.medi
 
 =head1 VERSION
 
-Version 0.38
+Version 0.41
 
 =cut
 
-our $VERSION  = "0.38";
+our $VERSION  = "0.41";
 
 =head1 SYNOPSIS
 
@@ -343,7 +343,7 @@ sub api {
     unless $self->{config}->{api_url};
 
   my $retries = $self->{config}->{retries};
-  my $maxlagretries = $self->{config}->{max_lag_retries};
+  my $maxlagretries = 1;
 
   $self->_encode_hashref_utf8($query, $options->{skip_encoding});
   $query->{maxlag} = $self->{config}->{max_lag} if defined $self->{config}->{max_lag}; 
@@ -425,13 +425,11 @@ sub api {
 
     # check lag and wait
     if (ref($ref) eq 'HASH' && exists $ref->{error} && $ref->{error}->{code} eq 'maxlag' ) {
-      $ref->{'error'}->{'info'} =~ /: (\d+) seconds lagged/;
-      my $lag = $1;
-      if ($maxlagretries == 0) {
-        return $self->_error(ERR_API,"Server has reported lag above the configure max_lag value of " . $self->{config}->{max_lag} . " value after " .($maxlagretries+1)." attempt(s). Last reported lag was - ". $ref->{'error'}->{'info'})
+      if ($maxlagretries == $self->{config}->{max_lag_retries}) {
+        return $self->_error(ERR_API,"Server has reported lag above the configured max_lag value of " . $self->{config}->{max_lag} . " value after " . $self->{config}->{max_lag_retries} . " attempt(s). Last reported lag was - ". $ref->{'error'}->{'info'})
       } else {
         sleep $self->{config}->{max_lag_delay};
-        $maxlagretries-- if $maxlagretries > 0;
+        $maxlagretries++ if $maxlagretries < $self->{config}->{max_lag_retries};
         # redo the request
         next;
       }
@@ -486,6 +484,12 @@ Currently
 
 =item * (Un)block users (Mediawiki >= 1.12 )
 
+=item * (Un)watch a page (Mediawiki >= 1.18 )
+
+=item * Email user (Mediawiki >= 1.14 )
+
+=item * Patrol changes (Mediawiki >= 1.14 )
+
 =back
 
 are supported via this call. Use this call to edit pages without having to worry about getting an edit token from the API first. The function will cache edit tokens to speed up future edits (Except for rollback edits, which are not cachable).
@@ -647,6 +651,8 @@ sub list {
 
   $options->{max} = 0 if ( !defined $options->{max} );
 
+  $query->{'rawcontinue'} = '';
+
   my $continue = 0;
   my $count = 0;
   do {
@@ -789,7 +795,7 @@ sub download {
 
   my $url = @{ $pageref->{imageinfo} }[0]->{url};
 
-  unless ( $url =~ /^http\:\/\// ) {
+  unless ( $url =~ /^https?\:\/\// ) {
     return $self->_error(ERR_PARAMS,'The API returned a relative path. You need to configure the url where files are stored in {config}->{files_url}')
       unless ( defined $self->{config}->{files_url} );
     $url = $self->{config}->{files_url} . $url;
@@ -856,71 +862,70 @@ sub _make_querystring {
 # gets a token for a specified parameter and sets it in the query for the call
 sub _get_set_tokens {
   my ($self, $query) = @_;
-  my ($prop, $title, $token);
-  
-  my $action = $query->{action};
 
-  if ( $action eq 'move' ) {
-    $title = $query->{from};
-  } else {
-    $title = $query->{title};
-  }
-
-  if ( $action eq 'upload' ) {
-    $action = 'edit';
-    $title = $query->{filename};
-  }
-
-  # check if we have a cached token.
-  if ( exists( $self->{config}->{tokens}->{$action} ) ) {
-    $query->{token} = $self->{config}->{tokens}->{$action};
+  # Note that although currently many of the tokens are equivalent, we cache them separately
+  # in case this was to change.
+  if ( exists( $self->{config}->{tokens}->{$query->{action}} ) ) {
+    $query->{token} = $self->{config}->{tokens}->{$query->{action}};
     return 1;
   }
 
-  # if we are doing an import, get the edit token using Main_Page as API docs suggest.
-  if ( $action eq 'import' ) {
-    # if a title is supplied use that page to get the edit token instead of Main_Page
-    if ( defined $query->{title} ) {
-      $title = $query->{title};
-    } else {
-      $title = "Main_Page";
-    }
+  # set the token actions and title depending on edit action
+  my $tokaction = $query->{action};
+  my $title = $query->{title};
+  if ( $query->{action} eq 'move' ) {
+    $title = $query->{from};
+  } elsif ( $query->{action} eq 'upload' ) {
+    $tokaction = 'edit';
+    $title = $query->{filename};
+  } elsif ( $query->{action} eq 'emailuser' ) {
+    $tokaction = 'email';
+    $title = 'User:' . $query->{target};
+  } elsif ( $query->{action} eq 'import' ) {
+    # if title was not specified get the import token using Main_Page as API docs suggest.
+    $title = "Main_Page" unless defined $query->{title};
   }
 
-  # set the properties we want to extract based on the action
-  if ( $action eq 'rollback' ) {
-    $prop = 'revisions'; 
-  } else {
-    $prop = 'info';
+  # set the properties and token name we want to extract based on the action
+  # may merge into above ifdef depending on other token implementations
+  my $prop = 'info';
+  my $token = 'intoken';
+  my $tokquery = { action => 'query', prop => $prop, $token => $tokaction, titles => $title };  
+  if ( $tokaction eq 'rollback' ) {
+    $prop = 'revisions';
+    $token = 'rvtoken';
+  } elsif ( $tokaction eq 'patrol' ) {
+    $tokquery = { action => 'query', list => 'recentchanges', rclimit => '1', rctoken => $tokaction };
   }
 
-  $token = 'intoken';
-  $token = 'rvtoken' if ( $action eq 'rollback' );
-
-  return undef unless ( my $ref = $self->api( { action => 'query', prop => $prop, $token => $action, titles => $title } ) );
+  my $ref;
+  $ref = $self->api( $tokquery );
+  return undef unless $ref;
 
   my ($pageid, $pageref) = each %{ $ref->{query}->{pages} };
 
   # if the page doesn't exist and we aren't editing/creating a new page then return an error
-  if ( defined $pageref->{missing} && $action ne 'edit' && $action ne 'import' ) {
-    return $self->_error( ERR_EDIT, "Unable to $action page '$title'. Page does not exist.") 
+  if ( defined $pageref->{missing} && $tokaction ne 'edit' && $tokaction ne 'import' ) {
+    return $self->_error( ERR_EDIT, "Unable to $tokaction page '$title'. Page does not exist.") 
   }
 
-  if ( $action eq 'rollback' ) {
-    $query->{token} = @{ $pageref->{revisions} }[0]->{$action.'token'};
+  if ( $query->{action} eq 'rollback' ) {
+    $query->{token} = @{ $pageref->{revisions} }[0]->{$tokaction.'token'};
     my $lastuser = @{ $pageref->{revisions} }[0]->{user};
     $query->{user} = @{ $pageref->{revisions} }[0]->{user} unless defined $query->{user};
     return $self->_error( ERR_EDIT, "Unable to rollback edits from user '$query->{user}' for page '$title'. Last edit was made by user $lastuser" ) if ( $query->{user} ne $lastuser );
+  } elsif ( $tokaction eq 'patrol' ) {
+    $query->{token} = @{$ref->{query}->{recentchanges}}[0]->{patroltoken};
   } else {
-    $query->{token} = $pageref->{$action.'token'};
+    $query->{token} = $pageref->{$tokaction.'token'};
   }
 
-  return $self->_error( ERR_EDIT, "Unable to get an edit token for action '$action'." ) unless ( defined $query->{token} );
+  return $self->_error( ERR_EDIT, "Unable to get an edit token for action '$tokaction'." ) unless ( defined $query->{token} );
 
   # cache the token. rollback tokens are specific for the page name and last edited user so can not be cached. Note that although currently many of the tokens
   # are equivalent, we cache them separately in case this was to change.
-  if ( $action ne 'rollback' ) {
-    $self->{config}->{tokens}->{$action} = $query->{token};
+  if ( $query->{action} ne 'rollback' ) {
+    $self->{config}->{tokens}->{$query->{action}} = $query->{token};
   }
 
   return 1;
